/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    CDT.java
 *    This class is strongly based on the REPTree class which extends (the 
 *    author of the REPTree class is Eibe Frank, eibe@cs.waikato.ac.nz ). As a 
 *    result, a significant part of the code is a modified version of the 
 *    original code. Futhermore, extending an inner class from a subclass, 
 *    without using a factory method in the superclass, forces us to copy 
 *    some methods as they were. All the methods indicate whether they are an
 *    original copied method or a modified/new version.
 */

package weka.classifiers.trees;

import weka.classifiers.Sourcable;
import weka.classifiers.rules.ZeroR;
import weka.core.AdditionalMeasureProducer;
import weka.core.Drawable;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.RevisionUtils;
import weka.core.Utils;
import weka.core.WeightedInstancesHandler;
import weka.core.ImpreciseEntropyContingencyTables; 


import java.io.Serializable;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Random;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

/*--------------------------------------------------------------------------*/
/**
 <!-- globalinfo-start -->
 * <p>Decision tree learner based on imprecise probabilities and 
      uncertainty measures. See:</p>
      <pre> J. Abellán, S. Moral
            Building classification trees using the total uncertainty criterion
            Int. J. Intell. Syst., 18 (12) (2003), pp. 1215–1225
      </pre>
   <p>It uses reduced-error pruning (with backfitting) and sorted values for 
      numeric attributes. Missing values are treated like C4.5
 * </p>
 <!-- globalinfo-end -->
 *
 <!-- options-start -->
 * <p>Valid options are: </p>
 * 
 * <pre>
 * -M &lt;minimum number of instances&gt;
 *  Set minimum number of instances per leaf (default 2).
 * </pre>
 * 
 * <pre>
 * -V &lt;minimum variance for split&gt;
 *  Set minimum numeric class variance proportion
 *  of train variance for split (default 1e-3).
 * </pre>
 * 
 * <pre>
 * -N &lt;number of folds&gt;
 *  Number of folds for reduced error pruning (default 3).
 * </pre>
 * 
 * <pre>
 * -S &lt;seed&gt;
 *  Seed for random data shuffling (default 1).
 * </pre>
 * 
 * <pre>
 * -P
 *  No pruning.
 * </pre>
 * 
 * <pre>
 * -L &lt;depth&gt;
 *  Maximum tree depth (default -1, no maximum)
 * </pre>
 * <pre> 
 * -SIDM &lt;s&gt;
 *  S parameter used in the Imprecise Dirichlet Model to obtain imprecise 
 *  probabilities as intervals. By default s=1. 
 * </pre>
 * <pre> 
 * -KTHRootAttribute &lt;k&gt;
 *  This value is used to fix the root node of the tree (using k-th highest 
 *  variable of a rank based on Information Gain using imprecise probabilities). 
 *  By default is k=1.
 * </pre>
 <!-- options-end -->
 *
 * @author Eibe Frank (eibe@cs.waikato.ac.nz) ( code from REPTRee.java )
 * @author Joaquín Abellán (jabellan@decsai.ugr.es)
 * @author Andrés Masegosa (andrew@decsai.ugr.es)
 * @author Javier G. Castellano (fjgc@decsai.ugr.es)
 * @version $Revision: 3 $ 
 */
public class CDT 
  extends REPTree 
  implements OptionHandler, WeightedInstancesHandler, Drawable, 
	     AdditionalMeasureProducer, Sourcable {

  /** for serialization */
  static final long serialVersionUID = -8562443428621539458L;
  
  /** s value of the Imprecise Dirichlet Model that is used to obtain imprecise
   * probabilities. */
  double m_SValue=1.0;
  
  
  /**This property is used to fix the root node of the tree (using k-th highest 
   * variable of a rank based on Information Gain using imprecise probabilities). 
   * It is mainly used for the IRNV (Information Root Node Variation) ensemble. 
   * The default value is k=1. More info: 
   *    Joaquín Abellán, Andrés R. Masegosa 
   *    An ensemble method using credal decision trees, 
   *    European Journal of Operational Research, 205(1), 218-226, 2010.
   *    doi: 10.1016/j.ejor.2009.12.003.*/
  int m_KThRootAttribute=1;
  
  /** The Tree object. Redefined here to use the extended inner class  */
  protected CDT.Tree m_Tree = null;
  
  /*--------------------------------------------------------------------------*/
  /**
   * Returns a string describing classifier
   * @return a description suitable for
   * displaying in the explorer/experimenter gui
   */
  @Override
  public String globalInfo() {

    return  "Decision tree learner based on imprecise probabilities and "
      +"uncertainty measures. See:\n"
      +"\tJ. Abellán, S. Moral\n" 
      +"\tBuilding classification trees using the total uncertainty criterion\n" 
      +"\tInt. J. Intell. Syst., 18 (12) (2003), pp. 1215–1225\n\n"
   
      +"It uses reduced-error pruning (with backfitting) and sorted values for "
      +"numeric attributes. Missing values are treated like C4.5";
  }

  /*--------------------------------------------------------------------------*/
  /*--------------------------------------------------------------------------*/
  /** An inner class for building and storing the tree structure.
  
    Important Note:Extending an inner class from a subclass, without using a 
    factory method in the superclass, forces us to copy some methods as they 
    were in the superclass. All the methods indicate whether they are an
    original copied method or an extended/new version.
  */
                         
  protected class Tree extends REPTree.Tree
    implements Serializable {
    
    /** for serialization */
    static final long serialVersionUID = -163542346754537935L;
    
    /** The subtrees of this tree. */ 
    protected CDT.Tree[] m_Successors;
    
    /*--------------------------------------------------------------------------*/
    /**
     * Computes class distribution of an instance using the tree. 
     * Note: Original copied method.
     * 
     * @param instance the instance to compute the distribution for
     * @return the distribution
     * @throws Exception if computation fails
     */
    @Override
    protected double[] distributionForInstance(Instance instance) 
    throws Exception {

      double[] returnedDist = null;
      
      if (m_Attribute > -1) {
	
	// Node is not a leaf
	if (instance.isMissing(m_Attribute)) {
	  
	  // Value is missing
	  returnedDist = new double[m_Info.numClasses()];

	  // Split instance up
	  for (int i = 0; i < m_Successors.length; i++) {
	    double[] help = m_Successors[i].distributionForInstance(instance);
	    if (help != null) {
	      for (int j = 0; j < help.length; j++) {
		returnedDist[j] += m_Prop[i] * help[j];
	      }
	    }
	  }
	} else if (m_Info.attribute(m_Attribute).isNominal()) {
	  
	  // For nominal attributes
	  returnedDist =  m_Successors[(int)instance.value(m_Attribute)].
	    distributionForInstance(instance);
	} else {
	  
	  // For numeric attributes
	  if (instance.value(m_Attribute) < m_SplitPoint) {
	    returnedDist = 
	      m_Successors[0].distributionForInstance(instance);
	  } else {
	    returnedDist = 
	      m_Successors[1].distributionForInstance(instance);
	  }
	}
      }
      if ((m_Attribute == -1) || (returnedDist == null)) {

        // Node is a leaf or successor is empty
        if (m_ClassProbs == null) {
          return m_ClassProbs;
        }
        return m_ClassProbs.clone();
      } else {
        return returnedDist;
      }
    }

    /*------------------------------------------------------------------------*/
    /**
     * Recursively outputs the tree.
     * Note: Original copied method.
     * 
     * @param level the current level
     * @param parent the current parent
     * @return the generated substree
     */
    protected String toString(int level, Tree parent) {

      try {
	StringBuilder text = new StringBuilder();
      
	if (m_Attribute == -1) {
	
	  // Output leaf info
	  return leafString(parent);
	} else if (m_Info.attribute(m_Attribute).isNominal()) {
	
	  // For nominal attributes
	  for (int i = 0; i < m_Successors.length; i++) {
	    text.append("\n");
	    for (int j = 0; j < level; j++) {
	      text.append("|   ");
	    }
	    text.append(m_Info.attribute(m_Attribute).name())
                .append(" = ")
                .append(m_Info.attribute(m_Attribute).value(i))
	        .append(m_Successors[i].toString(level + 1, this));
	  }
	} else {
	
	  // For numeric attributes
	  text.append("\n");
	  for (int j = 0; j < level; j++) {
	    text.append("|   ");
	  }
	  text.append(m_Info.attribute(m_Attribute).name())
              .append(" < ").append(Utils.doubleToString(m_SplitPoint, 2))
	      .append(m_Successors[0].toString(level + 1, this))
	      .append("\n");
	  for (int j = 0; j < level; j++) {
	    text.append("|   ");
	  }
	  text.append(m_Info.attribute(m_Attribute).name())
              .append(" >= ").append(Utils.doubleToString(m_SplitPoint, 2))
	      .append(m_Successors[1].toString(level + 1, this));
	}
      
	return text.toString();
      } catch (Exception e) {
        Logger.getLogger(CDT.class.getName()).log(Level.WARNING, null, e);                           
	return "WARNING: Decision tree: tree can't be printed";
      }
    }   
    
    
    /*------------------------------------------------------------------------*/
    /**
     * Outputs one node for graph.
     * 
     * Note: original modified method.
     * 
     * @param text the buffer to append the output to
     * @param num the current node id
     * @param parent the parent of the nodes
     * @return the next node id
     * @throws Exception if something goes wrong
     */
    protected int toGraph(StringBuffer text, int num, Tree parent)
      throws Exception {

      num++;
      if (m_Attribute == -1) {
        text.append("N")
                .append(Integer.toHexString(Tree.this.hashCode()))
                .append(" [label=\"")
                .append(num)
                .append(Utils.backQuoteChars(leafString(parent)))
                .append("\"shape=box]\n");
      } else {
        text.append("N").append(Integer.toHexString(Tree.this.hashCode()))
                .append(" [label=\"")
                .append(num)
                .append(": ")
                .append(Utils.backQuoteChars(m_Info.attribute(m_Attribute).name()))
                .append("\"]\n");
        for (int i = 0; i < m_Successors.length; i++) {
          text.append("N")
                  .append(Integer.toHexString(Tree.this.hashCode()))
                  .append("->N")
                  .append(Integer.toHexString(m_Successors[i].hashCode()))
                  .append(" [label=\"");
          if (m_Info.attribute(m_Attribute).isNumeric()) {
            if (i == 0) {
              text.append(" < ").append(Utils.doubleToString(m_SplitPoint, 2));
            } else {
              text.append(" >= ").append(Utils.doubleToString(m_SplitPoint, 2));
            }
          } else {
            text.append(" = ")
                .append(Utils.backQuoteChars(m_Info.attribute(m_Attribute).value(i)));
          }
          text.append("\"]\n");
          num = m_Successors[i].toGraph(text, num, this);
        }
      }

      return num;
    }
    /*------------------------------------------------------------------------*/
    /**
     * Recursively generates a tree.
     * Note: Inherited modified method.
     * 
     * @param sortedIndices the sorted indices of the instances
     * @param weights the weights of the instances
     * @param data the data to work with
     * @param totalWeight
     * @param classProbs the class probabilities
     * @param header the header of the data
     * @param minNum the minimum number of instances in a leaf
     * @param minVariance
     * @param depth the current depth of the tree
     * @param maxDepth the maximum allowed depth of the tree
     * @throws Exception if generation fails
     */
    @Override
    protected void buildTree(int[][][] sortedIndices, double[][][] weights,
      Instances data, double totalWeight, double[] classProbs,
      Instances header, double minNum, double minVariance, int depth,
      int maxDepth) throws Exception {

      // Store structure of dataset, set minimum number of instances
      // and make space for potential info from pruning data
      m_Info = header;
      if (data.classAttribute().isNumeric()) {
        m_HoldOutDist = new double[2];
      } else {
        m_HoldOutDist = new double[data.numClasses()];
      }

      // Make leaf if there are no training instances
      int helpIndex = 0;
      if (data.classIndex() == 0) {
        helpIndex = 1;
      }
      if (sortedIndices[0][helpIndex].length == 0) {
        if (data.classAttribute().isNumeric()) {
          m_Distribution = new double[2];
        } else {
          m_Distribution = new double[data.numClasses()];
        }
        m_ClassProbs = null;
        sortedIndices[0] = null;
        weights[0] = null;
        return;
      }

      double priorVar = 0;
      if (data.classAttribute().isNumeric()) {

        // Compute prior variance
        double totalSum = 0, totalSumSquared = 0, totalSumOfWeights = 0;
        for (int i = 0; i < sortedIndices[0][helpIndex].length; i++) {
          Instance inst = data.instance(sortedIndices[0][helpIndex][i]);
          totalSum += inst.classValue() * weights[0][helpIndex][i];
          totalSumSquared += inst.classValue() * inst.classValue()
            * weights[0][helpIndex][i];
          totalSumOfWeights += weights[0][helpIndex][i];
        }
        priorVar = singleVariance(totalSum, totalSumSquared, totalSumOfWeights);
      }

      // Check if node doesn't contain enough instances, is pure
      // or the maximum tree depth is reached
      m_ClassProbs = new double[classProbs.length];
      System.arraycopy(classProbs, 0, m_ClassProbs, 0, classProbs.length);
      if ((totalWeight < (2 * minNum))
        ||

        // Nominal case
        (data.classAttribute().isNominal() && Utils.eq(
          m_ClassProbs[Utils.maxIndex(m_ClassProbs)], Utils.sum(m_ClassProbs)))
        ||

        // Numeric case
        (data.classAttribute().isNumeric() && ((priorVar / totalWeight) < minVariance))
        ||

        // Check tree depth
        ((maxDepth >= 0) && (depth >= maxDepth))) {

        // Make leaf
        m_Attribute = -1;
        if (data.classAttribute().isNominal()) {

          // Nominal case
          m_Distribution = new double[m_ClassProbs.length];
          System.arraycopy(m_ClassProbs, 0, m_Distribution, 0, 
                            m_ClassProbs.length);
          doSmoothing();
          Utils.normalize(m_ClassProbs);
        } else {

          // Numeric case
          m_Distribution = new double[2];
          m_Distribution[0] = priorVar;
          m_Distribution[1] = totalWeight;
        }
        sortedIndices[0] = null;
        weights[0] = null;
        return;
      }

      // Compute class distributions and value of splitting
      // criterion for each attribute
      double[] vals = new double[data.numAttributes()];
      double[][][] dists = new double[data.numAttributes()][0][0];
      double[][] props = new double[data.numAttributes()][0];
      double[][] totalSubsetWeights = new double[data.numAttributes()][0];
      double[] splits = new double[data.numAttributes()];
      if (data.classAttribute().isNominal()) {

        // Nominal case
        for (int i = 0; i < data.numAttributes(); i++) {
          if (i != data.classIndex()) {
            splits[i] = distribution(props, dists, i, sortedIndices[0][i],
              weights[0][i], totalSubsetWeights, data);
            vals[i] = gain(dists[i], priorVal(dists[i]));
          }
        }
      } else {

        // Numeric case
        for (int i = 0; i < data.numAttributes(); i++) {
          if (i != data.classIndex()) {
            splits[i] = numericDistribution(props, dists, i,
              sortedIndices[0][i], weights[0][i], totalSubsetWeights, data,
              vals);
          }
        }
      }

      // Find best attribute
      if (depth==0 && getKTHRootAttribute()>0){ 
             
         //For the root get the k-th node if it's set different than the best
         m_Attribute = Utils.sort(vals)[vals.length-getKTHRootAttribute()];
        
      }else{
          m_Attribute = Utils.maxIndex(vals);
      }
      int numAttVals = dists[m_Attribute].length;

      // Check if there are at least two subsets with
      // required minimum number of instances
      int count = 0;
      for (int i = 0; i < numAttVals; i++) {
        if (totalSubsetWeights[m_Attribute][i] >= minNum) {
          count++;
        }
        if (count > 1) {
          break;
        }
      }

      // Any useful split found?
      if (Utils.gr(vals[m_Attribute], 0) && (count > 1)) {

        // Set split point, proportions, and temp arrays
        m_SplitPoint = splits[m_Attribute];
        m_Prop = props[m_Attribute];
        double[][] attSubsetDists = dists[m_Attribute];
        double[] attTotalSubsetWeights = totalSubsetWeights[m_Attribute];

        // Release some memory before proceeding further
        vals = null;
        dists = null;
        props = null;
        totalSubsetWeights = null;
        splits = null;

        // Split data
        int[][][][] subsetIndices = new int[numAttVals][1][data.numAttributes()][0];
        double[][][][] subsetWeights = new double[numAttVals][1][data
          .numAttributes()][0];
        splitData(subsetIndices, subsetWeights, m_Attribute, m_SplitPoint,
          sortedIndices[0], weights[0], data);

        // Release memory
        sortedIndices[0] = null;
        weights[0] = null;

        // Build successors
        m_Successors = new CDT.Tree[numAttVals];
        for (int i = 0; i < numAttVals; i++) {
          m_Successors[i] = new CDT.Tree();
          m_Successors[i].buildTree(subsetIndices[i], subsetWeights[i], data,
            attTotalSubsetWeights[i], attSubsetDists[i], header, minNum,
            minVariance, depth + 1, maxDepth);

          // Release as much memory as we can
          attSubsetDists[i] = null;
        }
      } else {

        // Make leaf
        m_Attribute = -1;
        sortedIndices[0] = null;
        weights[0] = null;
      }

      // Normalize class counts
      if (data.classAttribute().isNominal()) {
        m_Distribution = new double[m_ClassProbs.length];
        System.arraycopy(m_ClassProbs, 0, m_Distribution, 0, m_ClassProbs.length);
        doSmoothing();
        Utils.normalize(m_ClassProbs);
      } else {
        m_Distribution = new double[2];
        m_Distribution[0] = priorVar;
        m_Distribution[1] = totalWeight;
      }
    }

    /*------------------------------------------------------------------------*/
    /**
     * Computes size of the tree.
     * Note: Original modified method.
     * 
     * @return the number of nodes
     */
   @Override
    protected int numNodes() {
    
      if (m_Attribute == -1) {
	return 1;
      } else {
	int size = 1;
          for (Tree m_Successor : m_Successors) {
              size += m_Successor.numNodes();
          }
	return size;
      }
    }

    /*------------------------------------------------------------------------*/
    /**
     * Computes value of splitting criterion before split.
     * Note: modified method.
     * 
     * @param dist
     * @return the splitting criterion
     */
    @Override
    protected double priorVal(double[][] dist) {

      return ImpreciseEntropyContingencyTables.impreciseEntropyOverColumns(dist,CDT.this.getSValue());
    }

    /*------------------------------------------------------------------------*/
    /**
     * Computes value of splitting criterion after split.
     * Note: modified method.
     * 
     * @param dist
     * @param priorVal the splitting criterion
     * @return the gain after splitting
     */
    @Override
    protected double gain(double[][] dist, double priorVal) {

      return priorVal - ImpreciseEntropyContingencyTables.impreciseEntropyConditionedOnRows(dist,CDT.this.getSValue());
    }

    /*------------------------------------------------------------------------*/
    /**
     * Prunes the tree using the hold-out data (bottom-up).
     * Note: Original modified method.
     * 
     * @return the error
     * @throws Exception if pruning fails for some reason
     */
    @Override
    protected double reducedErrorPrune() throws Exception {

      // Is node leaf ? 
      if (m_Attribute == -1) {
	return m_HoldOutError;
      }

      // Prune all sub trees
      double errorTree = 0;
        for (Tree m_Successor : m_Successors) {
            errorTree += m_Successor.reducedErrorPrune();
        }

      // Replace sub tree with leaf if error doesn't get worse
      if (errorTree >= m_HoldOutError) {
	m_Attribute = -1;
	m_Successors = null;
	return m_HoldOutError;
      } else {
	return errorTree;
      }
    }

    /*------------------------------------------------------------------------*/
    /**
     * Inserts hold-out set into tree.
     * Note: Original copied method.
     * 
     * @param data the data to insert
     * @throws Exception if something goes wrong
     */
    @Override
    protected void insertHoldOutSet(Instances data) throws Exception {

      for (int i = 0; i < data.numInstances(); i++) {
	insertHoldOutInstance(data.instance(i), data.instance(i).weight(),
			      this);
      }
    }

    /*------------------------------------------------------------------------*/
    /**
     * Inserts an instance from the hold-out set into the tree.
     * Note: Original copied method.
     * 
     * @param inst the instance to insert
     * @param weight the weight of the instance
     * @param parent the parent of the node
     * @throws Exception if insertion fails
     */
    protected void insertHoldOutInstance(Instance inst, double weight,
      Tree parent) throws Exception {

      // Insert instance into hold-out class distribution
      if (inst.classAttribute().isNominal()) {

        // Nominal case
        m_HoldOutDist[(int) inst.classValue()] += weight;
        int predictedClass;
        if (m_ClassProbs == null) {
          predictedClass = Utils.maxIndex(parent.m_ClassProbs);
        } else {
          predictedClass = Utils.maxIndex(m_ClassProbs);
        }
        if (predictedClass != (int) inst.classValue()) {
          m_HoldOutError += weight;
        }
      } else {

        // Numeric case
        m_HoldOutDist[0] += weight;
        m_HoldOutDist[1] += weight * inst.classValue();
        double diff;
        if (m_ClassProbs == null) {
          diff = parent.m_ClassProbs[0] - inst.classValue();
        } else {
          diff = m_ClassProbs[0] - inst.classValue();
        }
        m_HoldOutError += diff * diff * weight;
      }

      // The process is recursive
      if (m_Attribute != -1) {

        // If node is not a leaf
        if (inst.isMissing(m_Attribute)) {

          // Distribute instance
          for (int i = 0; i < m_Successors.length; i++) {
            if (m_Prop[i] > 0) {
              m_Successors[i].insertHoldOutInstance(inst, weight * m_Prop[i],
                this);
            }
          }
        } else {

          if (m_Info.attribute(m_Attribute).isNominal()) {

            // Treat nominal attributes
            m_Successors[(int) inst.value(m_Attribute)].insertHoldOutInstance(
              inst, weight, this);
          } else {

            // Treat numeric attributes
            if (inst.value(m_Attribute) < m_SplitPoint) {
              m_Successors[0].insertHoldOutInstance(inst, weight, this);
            } else {
              m_Successors[1].insertHoldOutInstance(inst, weight, this);
            }
          }
        }
      }
    }

    /*------------------------------------------------------------------------*/
    /**
     * Backfits data from holdout set.
     * Note: Original copied method.
     * 
     * @throws Exception if insertion fails
     */
    @Override
    protected void backfitHoldOutSet() throws Exception {

      // Insert instance into hold-out class distribution
      if (m_Info.classAttribute().isNominal()) {

        // Nominal case
        if (m_ClassProbs == null) {
          m_ClassProbs = new double[m_Info.numClasses()];
        }
        System.arraycopy(m_Distribution, 0, m_ClassProbs, 0,
          m_Info.numClasses());
        for (int i = 0; i < m_HoldOutDist.length; i++) {
          m_ClassProbs[i] += m_HoldOutDist[i];
        }
        if (Utils.sum(m_ClassProbs) > 0) {
          doSmoothing();
          Utils.normalize(m_ClassProbs);
        } else {
          m_ClassProbs = null;
        }
      } else {

        // Numeric case
        double sumOfWeightsTrainAndHoldout = m_Distribution[1]
          + m_HoldOutDist[0];
        if (sumOfWeightsTrainAndHoldout <= 0) {
          return;
        }
        if (m_ClassProbs == null) {
          m_ClassProbs = new double[1];
        } else {
          m_ClassProbs[0] *= m_Distribution[1];
        }
        m_ClassProbs[0] += m_HoldOutDist[1];
        m_ClassProbs[0] /= sumOfWeightsTrainAndHoldout;
      }

      // The process is recursive
      if (m_Attribute != -1) {
        for (CDT.Tree m_Successor : m_Successors) {
          m_Successor.backfitHoldOutSet();
        }
      }
    }
    
  }//End inner class    
  /*--------------------------------------------------------------------------*/
  /*--------------------------------------------------------------------------*/
  
  /*--------------------------------------------------------------------------*/
  /**
     * This method gets the s parameter used in the Imprecise Dirichlet model 
     * to obtain imprecise probabilities. 
     * @return s parameter of the IDM
     */
    public double getSValue() {
        return m_SValue;
    }

    /*------------------------------------------------------------------------*/
    /**
     * Set the s parameter used in the Imprecise Dirichlet model to obtain 
     * imprecise probabilities. 
     * @param m_SValue s parameter of the IDM
     */
    public void setSValue(double m_SValue) {
        if (m_SValue<0) m_SValue=0; //sanity check
        this.m_SValue = m_SValue;
    }
    /*------------------------------------------------------------------------*/
    /**
     * Returns the tip text for this property
     * 
     * @return tip text for this property suitable for displaying in the
     *         explorer/experimenter gui
     */
    public String SValueTipText() {
      return "Parameter used in the Imprecise Dirichlet Model to obtain "+
              "imprecise probabilities as intervals. By default s=1.";
    }   
    
    /*------------------------------------------------------------------------*/
    /**
    * This method stores the position of the node in a rank to be used as root 
    * node in a decision tree (for the IRNV ensemble). 
    * @param k position in the rank to be used as root node. 
    */
    public void setKTHRootAttribute(int k) {
        if (k<1) k=1; //sanity check
        this.m_KThRootAttribute=k;
    }
    
    /*------------------------------------------------------------------------*/
    /**
    * This method gets the position of the node in a rank to be used as root 
    * node in a decision tree (for the IRNV ensemble). 
    * @return k position in the rank to be used as root node. 
    */
    public int getKTHRootAttribute() {
        return this.m_KThRootAttribute;
    }
    /*------------------------------------------------------------------------*/
    /**
     * Returns the tip text for this property
     * 
     * @return tip text for this property suitable for displaying in the
     *         explorer/experimenter gui
     */
    public String KTHRootAttributeTipText() {
      return "This property is used to fix the root node of the tree (using "+
              "the k-th highest variable of a rank based on Information Gain "+
              "using imprecise probabilities). It is mainly used for the IRNV"+
              " (Information Root Node Variation) ensemble. The default value"+
              " is k=1. More info: \n" +
              "\tJoaquín Abellán, Andrés R. Masegosa \n" +
              "\tAn ensemble method using credal decision trees, \n" +
              "\tEuropean Journal of Operational Research, 205(1), 218-226, 2010.\n" +
              "\tdoi: 10.1016/j.ejor.2009.12.003 ";
    }   
    
  /*--------------------------------------------------------------------------*/
  /**
   * Lists the command-line options for this classifier.
   * 
   * @return an enumeration over all commandline options
   */
  @Override
  public Enumeration listOptions() {

    ArrayList newArrayList = new ArrayList();

    //Add the s parameter to the options
    newArrayList.add(new Option(
      "\ts value used in the IDM (Imprecise Dirichlet Model)" + "(default 1).",
      "SIDM", 1,"-SIDM <value>"));
    
    //Add the Kth parameter to the options
    newArrayList.add(new Option(
      "\tk value used to fix the root node of the tree " + "(default 1).",
      "KTHRootAttribute", 1,"-KTHRootAttribute <value>"));
    
    //Add the options from the parent method
    newArrayList.addAll(Collections.list(super.listOptions()));
    
    return Collections.enumeration(newArrayList);
  } 

  /*------------------------------------------------------------------------*/
  /**
   * Gets options from this classifier.
   * 
   * @return the options for the current setup
   */
  @Override
  public String[] getOptions() {
    ArrayList<String> options = new ArrayList<>();
    
    //We get the "s" parameter for the IDM (Imprecise Dirichlet Model)
    options.add("-SIDM");
    options.add("" + getSValue());
    
    //We get the "k" value used to fix the root node of the tree
    options.add("-KTHRootAttribute");
    options.add("" + getKTHRootAttribute());
    
    //And add the parent options
    Collections.addAll(options, super.getOptions());

 
    return options.toArray(new String[0]);
  }
  
  /*------------------------------------------------------------------------*/
  /**
   * <p>Parses a given list of options. </p>
   * 
   * <!-- options-start -->
   * <pre>
   * -M &lt;minimum number of instances&gt;
   *  Set minimum number of instances per leaf (default 2).
   * </pre>
   * 
   * <pre>
   * -V &lt;minimum variance for split&gt;
   *  Set minimum numeric class variance proportion
   *  of train variance for split (default 1e-3).
   * </pre>
   * 
   * <pre>
   * -N &lt;number of folds&gt;
   *  Number of folds for reduced error pruning (default 3).
   * </pre>
   * 
   * <pre>
   * -S &lt;seed&gt;
   *  Seed for random data shuffling (default 1).
   * </pre>
   * 
   * <pre>
   * -P
   *  No pruning.
   * </pre>
   * 
   * <pre>
   * -L
   *  Maximum tree depth (default -1, no maximum)
   * </pre>
   * 
   * <pre>-SIDM
   *  S parameter used in the Imprecise Dirichlet Model to obtain imprecise 
   *    probabilities as intervals. By default s=1. </pre>
   * <pre>
   * -KTHRootAttribute 
   *    This value is used to fix the root node of the tree (using k-th highest 
   *    variable of a rank based on Information Gain using imprecise probabilities). 
   *    By default is k=1.
   * </pre>
   <!-- options-end -->
   * 
   * @param options the list of options as an array of strings
   * @throws Exception if an option is not supported
   */
  @Override
  public void setOptions(String[] options) throws Exception {
    
    String convertList = Utils.getOption("SIDM",options);
    if (convertList.length() != 0) {
        this.setSValue(Double.parseDouble(convertList));
    } else {
        this.setSValue(1);
    }
    
    convertList = Utils.getOption("KTHRootAttribute",options);
    if (convertList.length() != 0) {
        this.setKTHRootAttribute(Integer.parseInt(convertList));
    } else {
        this.setKTHRootAttribute(1);
    }
    
    super.setOptions(options);
    Utils.checkForRemainingOptions(options);
  }
  
 /*------------------------------------------------------------------------*/
 /**
   * Builds classifier.
   * Note: Inherited modified method.
   * 
   * @param data the data to train with
   * @throws Exception if building fails
   */
  @Override
  public void buildClassifier(Instances data) throws Exception {

    // can classifier handle the data?
    getCapabilities().testWithFail(data);

    // remove instances with missing class
    data = new Instances(data);
    data.deleteWithMissingClass();

    Random random = new Random(m_Seed);

    m_zeroR = null;
    if (data.numAttributes() == 1) {
      m_zeroR = new ZeroR();
      m_zeroR.buildClassifier(data);
      return;
    }

    // Randomize and stratify
    data.randomize(random);
    if (data.classAttribute().isNominal()) {
      data.stratify(m_NumFolds);
    }

    // Split data into training and pruning set
    Instances train;
    Instances prune = null;
    if (!m_NoPruning) {
      train = data.trainCV(m_NumFolds, 0, random);
      prune = data.testCV(m_NumFolds, 0);
    } else {
      train = data;
    }

    // Create array of sorted indices and weights
    int[][][] sortedIndices = new int[1][train.numAttributes()][0];
    double[][][] weights = new double[1][train.numAttributes()][0];
    double[] vals = new double[train.numInstances()];
    for (int j = 0; j < train.numAttributes(); j++) {
      if (j != train.classIndex()) {
        weights[0][j] = new double[train.numInstances()];
        if (train.attribute(j).isNominal()) {

          // Handling nominal attributes. Putting indices of
          // instances with missing values at the end.
          sortedIndices[0][j] = new int[train.numInstances()];
          int count = 0;
          for (int i = 0; i < train.numInstances(); i++) {
            Instance inst = train.instance(i);
            if (!inst.isMissing(j)) {
              sortedIndices[0][j][count] = i;
              weights[0][j][count] = inst.weight();
              count++;
            }
          }
          for (int i = 0; i < train.numInstances(); i++) {
            Instance inst = train.instance(i);
            if (inst.isMissing(j)) {
              sortedIndices[0][j][count] = i;
              weights[0][j][count] = inst.weight();
              count++;
            }
          }
        } else {

          // Sorted indices are computed for numeric attributes
          for (int i = 0; i < train.numInstances(); i++) {
            Instance inst = train.instance(i);
            vals[i] = inst.value(j);
          }
          sortedIndices[0][j] = Utils.sort(vals);
          for (int i = 0; i < train.numInstances(); i++) {
            weights[0][j][i] = train.instance(sortedIndices[0][j][i]).weight();
          }
        }
      }
    }

    // Compute initial class counts
    double[] classProbs = new double[train.numClasses()];
    double totalWeight = 0, totalSumSquared = 0;
    for (int i = 0; i < train.numInstances(); i++) {
      Instance inst = train.instance(i);
      if (data.classAttribute().isNominal()) {
        classProbs[(int) inst.classValue()] += inst.weight();
        totalWeight += inst.weight();
      } else {
        classProbs[0] += inst.classValue() * inst.weight();
        totalSumSquared += inst.classValue() * inst.classValue()
          * inst.weight();
        totalWeight += inst.weight();
      }
    }
    m_Tree = new CDT.Tree();
    double trainVariance = 0;
    if (data.classAttribute().isNumeric()) {
      trainVariance = m_Tree.singleVariance(classProbs[0], totalSumSquared,
        totalWeight) / totalWeight;
      classProbs[0] /= totalWeight;
    }

    // Build tree
    m_Tree.buildTree(sortedIndices, weights, train, totalWeight, classProbs,
      new Instances(train, 0), m_MinNum, m_MinVarianceProp * trainVariance, 0,
      m_MaxDepth);

    // Insert pruning data and perform reduced error pruning
    if (!m_NoPruning) {
      m_Tree.insertHoldOutSet(prune);
      m_Tree.reducedErrorPrune();
      m_Tree.backfitHoldOutSet();
    }
  }
  /*------------------------------------------------------------------------*/
  /**
   * Computes class distribution of an instance using the tree.
   * Note: Original copied method.
   * 
   * @param instance the instance to compute the distribution for
   * @return the computed class probabilities
   * @throws Exception if computation fails
   */
  @Override
  public double[] distributionForInstance(Instance instance) 
    throws Exception {
      
      if (m_zeroR != null) {
	return m_zeroR.distributionForInstance(instance);
      } else {
	return m_Tree.distributionForInstance(instance);
      }
  }

/*------------------------------------------------------------------------*/
/**
   * Computes size of the tree.
   * Note: Original copied method.
   * 
   * @return the number of nodes
   */
    @Override
  public int numNodes() {

    return m_Tree.numNodes();
  }

  /*------------------------------------------------------------------------*/
  /**
   * Returns the tree as if-then statements.
   * Note: modified inherited method.
   *
   * @param className the name for the generated class
   * @return the tree as a Java if-then type statement
   * @throws Exception if something goes wrong
   */
  @Override
  public String toSource(String className) 
    throws Exception {
     
    if (m_Tree == null) {
      throw new Exception("CDT: No model built yet.");
    } 
    StringBuffer [] source = m_Tree.toSource(className, m_Tree);
    return
    "class " + className + " {\n\n"
    +"  public static double classify(Object [] i)\n"
    +"    throws Exception {\n\n"
    +"    double p = Double.NaN;\n"
    + source[0]  // Assignment code
    +"    return p;\n"
    +"  }\n"
    + source[1]  // Support code
    +"}\n";
  }

  /*------------------------------------------------------------------------*/ 
  /**
   * Outputs the decision tree as a graph.
   * Note: modified inherited method.
   * 
   * @return the tree as a graph
   * @throws Exception if generation fails
   */
  @Override
  public String graph() throws Exception {

    if (m_Tree == null) {
      throw new Exception("CDT: No model built yet.");
    } 
    StringBuffer resultBuff = new StringBuffer();
    m_Tree.toGraph(resultBuff, 0, null);
    String result = "digraph Tree {\n" + "edge [style=bold]\n" + resultBuff.toString()
      + "\n}\n";
    return result;
  }
  
  /*------------------------------------------------------------------------*/
  /**
   * Outputs the decision tree.
   * Note: modified inherited method.
   * 
   * @return a string representation of the classifier 
   */
  @Override
  public String toString() {

    if (m_zeroR != null) {
      return "No attributes other than class. Using ZeroR.\n\n" + m_zeroR.toString();
    }
    if ((m_Tree == null)) {
      return "CDT: No model built yet.";
    } 
    return     
      "\nCDT\n============\n" + m_Tree.toString(0, null) + "\n" +
      "\nSize of the tree : " + numNodes();
  }
  
  /*------------------------------------------------------------------------*/
  /**
   * Returns the revision string.
   * 
   * @return the revision
   */
  @Override
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 3 $");
  }

  /*------------------------------------------------------------------------*/
  /**
   * Main method for this class.
   * 
   * @param argv the commandline options
   */
  public static void main(String[] argv) {
    runClassifier(new CDT(), argv);
  }

}//End class
