/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  ImpreciseEntropyContingencyTables.java
 *  Copyright (C) 2013 Joaquín Abellán (jabellan@decsai.ugr.es)
 *
 */
package weka.core;

import java.util.ArrayList;

/**
 * <!-- globalinfo-start -->
 * <p>This class inherits from the weka.core.ContingencyTables class of weka; 
 * The added functionality is for implementing a new entropy measure based on
 * imprecise probabilities and uncertainty measures.</p>
 * 
 * <p>For more info:</p>
 * <pre>
 *   J. Abellán, S. Moral
 *   Building classification trees using the total uncertainty criterion
 *   International Journal of Intelligent Systems, 18 (12) (2003), pp. 1215–1225 
 *   http://dx.doi.org/10.1002/int.10143
 * </pre>
 * 
 * @author Joaquín Abellán (jabellan@decsai.ugr.es)
 * @author Andrés Masegosa (andrew@decsai.ugr.es)
 * @author Javier G. Castellano (fjgc@decsai.ugr.es)
 * @version $Revision: 3.0 $
 * <!-- globalinfo-end -->
 */
public class ImpreciseEntropyContingencyTables extends ContingencyTables{

  /** for serialization */
  static final long serialVersionUID = 2689730616939923301L;
    
  /** Constant used to measure min differences*/
  static final double MINDIFF=0.00001;
 
  /** The natural logarithm of 2 */
  private static final double LOG2 = Math.log(2);
  
 
  /*--------------------------------------------------------------------------*/
  /**
   * This method distributes the mass of "s" to calculate the maximum entropy.
   * @param array the array
   * @param n the array effective size
   * @param svalue s parameter used for the IDM (Imprecise Dirichlet Model). 
   *                 The IDM is used to obtain imprecise probabilities as
   *                 intervals.
   * @return an array with the maximum entropy.
   */
  public static double[] sDistribution(double[] array,int n, double svalue){

        //Check if there is a non integer value.
        boolean nonintegers=false;
        for (int i=0;i<n;i++){
            if (((int)(array[i]))!=array[i]){
                nonintegers=true;
                break;
            }
        }

        if (nonintegers){

            double[] pl=new double[n];
            double[] pu=new double[n];

            double total=Utils.sum(array);
            ArrayList<Integer> indexes=new ArrayList();

            //we obtain the probability intervals using the IDM
            for (int i=0; i<n; i++){
                pl[i]=array[i]/(total+svalue);
                pu[i]=(array[i]+svalue)/(total+svalue);

                indexes.add(i);
            }

            return maxEntropy(pl, pu, indexes);
        }

        double mass;

        double[] l=new double[n];
        System.arraycopy(array,0,l,0,n);
        double min=l[0];
        int cont=0;
        for (int i=0;i<n;i++){
            if (min>l[i]) min=l[i];
        }
        for (int i=0;i<n;i++){
            if (min==l[i]) cont++;
        }
        if (svalue<=cont)
            mass=svalue;
        else
            mass=cont;
        for (int i=0;i<n;i++){
            if (min==l[i]) l[i]=l[i]+(mass/cont);
        }
        mass=mass-cont;
        if (mass>0.0001)
            return sDistribution(l,n,mass);
        else
            return l;

    }
  

    /*------------------------------------------------------------------------*/
    /**
     * Computes the entropy based on imprecise probabilities of the 
     * given array.
     *
     * @param array the array
     * @param svalue s parameter used for the IDM (Imprecise Dirchlet Model). 
     *                 The IDM is used to obtain imprecise probabilities as
     *                 intervals.
     * @return the entropy
     */
    public static double impreciseEntropy(double[] array, double svalue) {

      double returnValue = 0, sum = 0;
      double[] arraynew=sDistribution(array,array.length, svalue);

      for (int i = 0; i < arraynew.length; i++) {
        if (arraynew[i]>0)
          returnValue -= arraynew[i]*Utils.log2(arraynew[i]);
        sum += arraynew[i];
      }

      if (Utils.eq(sum, 0)) {
        return 0;
      } else {
        returnValue/=sum;
        return (returnValue + Utils.log2(sum));
      }
    }
 
    /*------------------------------------------------------------------------*/ 
    /**
     * This method computes the maximum entropy for intervals of probability 
     * distributions. More info in:
     *    Joaquin Abellan and Serafin Moral. 
     *    Maximum of entropy for credal sets. 
     *    Int. J. Uncertain. Fuzziness Knowl.-Based Syst. 11(5) , 587-597, 2003.
     *    doi: 10.1142/S021848850300234X 
     * @param pl an array with the lower values of the probability intervals.
     * @param pu another array with the upper values of the probability intervals.
     * @param indexes a list of indexes to use.
     * @return an array with the maximum entropy for each value.
     */
    private static double[] maxEntropy(double pl[], double pu[], 
            ArrayList<Integer> indexes){

        if (Math.abs(1-Utils.sum(pl))<0.001){ //if ((Sum(pl,m)<1.00001)&&(Sum(pl,m)>1.0-0.00001))   //SUM ONE
          double[] pmax=new double[pl.length];
          System.arraycopy(pl,0,pmax,0,pl.length);
          return pmax;
        } else {

           for (int i=0;i<indexes.size();i++){
               if (Math.abs(pl[indexes.get(i)]-pu[indexes.get(i)])<MINDIFF)
                   indexes.remove(i);
           }

           double sum=Utils.sum(pl); //the sum of 1 to n of the array pl
           int men=minIndex(pl,indexes); //index of the minimum value
           int sig=sigIndex(pl,men,indexes); //index of the second minor value
           int nme=nMin(pl,men,indexes); //num of indexes with the min value

           double val=pl[men];

           for (int i=0;i<pl.length;i++){
                  if (Math.abs(pl[i]-val)<MINDIFF){  //if ((pl[i]<val+0.00001)&&(pl[i]>val-0.00001)) {
                    if (!(sig==men)){
                       pl[i]=pl[i]+Math.min(Math.min(pu[i]-pl[i],(pl[sig]-val)),(1.0-sum)/(double)nme);
                    }else{
                       pl[i]=pl[i]+Math.min(Math.min(pu[i]-pl[i],(1.0-sum)/(double)nme),1);                       
                    }
                 }
           }
           return maxEntropy(pl,pu,indexes);
        }
    }  
    
    /*--------------------------------------------------------------------------*/
    /**
     * This method searches for the index of the minimum value of the array p.
     * @param p the array with the values.
     * @param indexes a list of indexes to use.
     * @return the index of the minimum value.
     */
    private static int minIndex(double p[], ArrayList<Integer> indexes){
        double min=Double.MAX_VALUE;
        int index=-1;
        for (int i=0; i<indexes.size(); i++)
            if (p[indexes.get(i)]<min){
                min=p[indexes.get(i)];
                index=indexes.get(i);
            }
        return index;
    }
    
    /*------------------------------------------------------------------------*/
    /**
     * This method computes the number of indexes that attains the minimum value
     * of the array.
     * @param p the array with the values.
     * @param indexlowest the index of the minimum value of the array. 
     * @param indexes a list of indexes to use.
     * @return the number of indexes that attains the minimum value
     */
    private static int nMin(double p[], int indexlowest, ArrayList<Integer> indexes){
        int nlowests=0;
        for (int i=0; i<indexes.size(); i++){
          if (Math.abs(p[indexes.get(i)]-p[indexlowest])<MINDIFF)//if ((p[i]>p[Menor(p,m)]-0.0000001)&&(p[i]<p[Menor(p,m)]+0.0000001)) 
              nlowests++;
        }
        return nlowests;
    }
    
    /*------------------------------------------------------------------------*/
    /**
     * This method searches for the index of the second minor value of the given 
     * array.
     * @param p an array with the values.
     * @param indexlowest the index of the minimum value of the array.
     * @param indexes a list of indexes to use.
     * @return the index of the second minor value of the array, it returns -1 
     *              if it is not exist.
     */
    private static int sigIndex(double p[],int indexlowest, ArrayList<Integer> indexes){
       double[] q = new double[indexes.size()];
       boolean different=false;
      for (int i=0; i<indexes.size(); i++){
         if (Math.abs(p[indexes.get(i)]-p[indexlowest])<MINDIFF){ //if (p[i]<p[Menor(p,m)]+0.00000001)
            q[i]=Double.MAX_VALUE;
         }else{
            different=true;
            q[i]=p[indexes.get(i)];
         }
      }
      if (different)
        return indexes.get(Utils.minIndex(q));
      else
        return indexlowest;
     }
      
    /*------------------------------------------------------------------------*/
    /**
     * Computes the columns' entropy (based on imprecise probabilities) for the 
     * given contingency table.
     *
     * @param matrix the contingency table
     * @param svalue s parameter used for the IDM (Imprecise Dirchlet Model). 
     *                 The IDM is used to obtain imprecise probabilities as
     *                 intervals.
     * @return the columns' entropy
     */
    public static double impreciseEntropyOverColumns(double[][] matrix, double svalue){

      double[] I_matrix=new double[matrix[0].length];


      for (int j = 0; j < matrix[0].length; j++){
          I_matrix[j] = 0;
          for (double[] matrix1 : matrix) {
              I_matrix[j] += matrix1[j];
          }
      }

      return impreciseEntropy(I_matrix,svalue);

    }
    
    /*------------------------------------------------------------------------*/
    /**
     * Computes conditional entropy (based on imprecise probabilities) of the 
     * columns given the rows.
     *
     * @param matrix the contingency table.
     * @param svalue s parameter used for the IDM (Imprecise Dirchlet Model). 
     *                 The IDM is used to obtain imprecise probabilities as
     *                 intervals.
     * @return the conditional entropy of the columns given the rows
     */
    public static double impreciseEntropyConditionedOnRows(double[][] matrix,double svalue) {

      double returnValue = 0, total = 0;
      double[][] I_matrix=new double[matrix.length][matrix[0].length];

      for (int i=0; i<matrix.length; i++)
         I_matrix[i]=sDistribution(matrix[i],matrix[0].length,svalue);

      for (int i = 0; i < matrix.length; i++) {
        double sumForRow = 0;
        double sumExactForRow = 0;
        double entropyForRow = 0;


        for (int j = 0; j < matrix[0].length; j++) {
              entropyForRow = entropyForRow + lnFunc(I_matrix[i][j]);
              sumForRow += I_matrix[i][j];

              sumExactForRow+=matrix[i][j];
              total+=matrix[i][j];
        }
        entropyForRow = (entropyForRow - lnFunc(sumForRow))/sumForRow;

        returnValue += sumExactForRow*entropyForRow;

      }
      if (Utils.eq(total, 0)) {
        return 0;
      }
      return -returnValue / (total * LOG2);
    }

    /*------------------------------------------------------------------------*/
    /**
     * Main method for testing this class.
     * @param args the commandline options
     */
    public static void main(String[] args) {

        //double[][] dist = {{0.010204449402759806, 2.131791585001831}, {0.03385018932347067, 0.0001620339186317366}};
        double svalue=1;
        double[][] dist = {{10, 3}, {4, 0}};

        System.out.println("Info-Gain: "+(ContingencyTables.entropyOverColumns(dist)-ContingencyTables.entropyConditionedOnRows(dist)));
        System.out.println("Info-Gain Ratio: "+(ContingencyTables.entropyOverColumns(dist)-ContingencyTables.entropyConditionedOnRows(dist))/ContingencyTables.entropyOverRows(dist));
        System.out.println("Imprecise Info-Gain: "+(ImpreciseEntropyContingencyTables.impreciseEntropyOverColumns(dist, svalue)-ImpreciseEntropyContingencyTables.impreciseEntropyConditionedOnRows(dist,svalue)));

        double[] cx=new double[4];//{0.010204449402759806, 2.131791585001831,0.03385018932347067, 0.0001620339186317366};
        double[] c= new double[2];//{0.010204449402759806+0.03385018932347067,2.131791585001831+0.0001620339186317366};
        double[] x= new double[2];//{0.010204449402759806+2.131791585001831,0.03385018932347067+0.0001620339186317366};

        cx[0]=dist[0][0];
        cx[1]=dist[0][1];
        cx[2]=dist[1][0];
        cx[3]=dist[1][1];

        c[0]=dist[0][0]+dist[1][0];
        c[1]=dist[0][1]+dist[1][1];

        x[0]=dist[0][0]+dist[0][1];
        x[1]=dist[1][0]+dist[1][1];

        System.out.println(ImpreciseEntropyContingencyTables.impreciseEntropy(c, svalue)+ImpreciseEntropyContingencyTables.impreciseEntropy(x, svalue)-ImpreciseEntropyContingencyTables.impreciseEntropy(cx, svalue));
    }

}









